import httpx

def req_wrap(funcname, *args, **kwargs) -> dict:
    func = getattr(httpx, funcname)
    res = func(*args, timeout=30, **kwargs)
    res.raise_for_status()
    return res.json() if kwargs["headers"]["Content-Type"] == "application/json" else dict(text=res.text)



import click
import sys
import logging
import asyncio

from boltz_client.boltz import BoltzConfig, BoltzClient

# disable tracebacks on exceptions
sys.tracebacklimit = 0

# better logging for cli
logger = logging.getLogger()
logger.setLevel(logging.INFO)


config = BoltzConfig(
    network="regtest",
    api_url="http://localhost:9001",
    mempool_url="http://localhost:8080",
    mempool_ws_url="ws://localhost:8080/api/v1/ws",
)


@click.group()
def command_group():
    """
Python CLI of boltz-client-python, enjoy submarine swapping. :)
Uses mempool.space for retrieving onchain data
"""

@click.command()
@click.argument('payment_request', type=str)
def create_swap(payment_request):
    """
    create a swap
    boltz will pay your invoice after you paid the onchain address

    SATS you want to swap, has to be the same as in PAYMENT_REQUEST
    PAYMENT_REQUEST with the same amount as specified in SATS
    """
    client = BoltzClient(config)
    refund_privkey_wif, swap = client.create_swap(payment_request)

    click.echo()
    click.echo(f"refund privkey in wif: {refund_privkey_wif}")
    click.echo(f"redeem_script_hex: {swap.redeemScript}")
    click.echo()
    click.echo(f"onchain address: {swap.address}")
    click.echo(f"expected amount: {swap.expectedAmount}")
    click.echo(f"bip21 address: {swap.bip21}")


@click.command()
@click.argument('privkey_wif', type=str)
@click.argument('lockup_address', type=str)
@click.argument('receive_address', type=str)
@click.argument('redeem_script_hex', type=str)
@click.argument('timeout_block_height', type=int)
def refund_swap(
    privkey_wif: str,
    lockup_address: str,
    receive_address: str,
    redeem_script_hex: str,
    timeout_block_height: int
):
    """
    refund a swap
    """
    client = BoltzClient(config)
    txid = client.refund_swap(
        privkey_wif,
        lockup_address,
        receive_address,
        redeem_script_hex,
        timeout_block_height,
    )
    click.echo(f"TXID: {txid}")


@click.command()
@click.argument('sats', type=int)
def create_reverse_swap(sats: int):
    """
    create a reverse swap
    """
    client = BoltzClient(config)
    claim_privkey_wif, preimage_hex, swap = client.create_reverse_swap(sats)

    click.echo("reverse swap created!")
    click.echo()
    click.echo(f"claim privkey in wif: {claim_privkey_wif}")
    click.echo(f"preimage hex: {preimage_hex}")
    click.echo(f"lockup_address: {swap.lockupAddress}")
    click.echo(f"redeem_script_hex: {swap.redeemScript}")
    click.echo()
    click.echo("invoice:")
    click.echo(swap.invoice)


@click.command()
@click.argument('receive_address', type=str)
@click.argument('sats', type=int)
@click.argument('zeroconf', type=bool, default=False)
def create_reverse_swap_and_claim(receive_address: str, sats: int, zeroconf: bool = False):
    """
    create a reverse swap and claim
    """
    client = BoltzClient(config)
    claim_privkey_wif, preimage_hex, swap = client.create_reverse_swap(sats)

    click.echo("reverse swap created!")
    click.echo()
    click.echo(f"claim privkey in wif: {claim_privkey_wif}")
    click.echo(f"preimage hex: {preimage_hex}")
    click.echo(f"lockup_address: {swap.lockupAddress}")
    click.echo(f"redeem_script_hex: {swap.redeemScript}")
    click.echo()
    click.echo("invoice:")
    click.echo(swap.invoice)
    click.echo()
    click.echo("waiting until you paid the invoice and")
    click.echo("boltz created the lockup transaction...")

    txid = asyncio.run(client.claim_reverse_swap(
        swap.lockupAddress,
        receive_address,
        claim_privkey_wif,
        preimage_hex,
        swap.redeemScript,
    ))

    click.echo(f"reverse swap claimed!")
    click.echo(f"TXID: {txid}")


@click.command()
@click.argument('lockup_address', type=str)
@click.argument('receive_address', type=str)
@click.argument('privkey_wif', type=str)
@click.argument('preimage_hex', type=str)
@click.argument('redeem_script_hex', type=str)
def claim_reverse_swap(
    lockup_address: str,
    receive_address: str,
    privkey_wif: str,
    preimage_hex: str,
    redeem_script_hex: str
):
    """
    claims a reverse swap
    """
    client = BoltzClient(config)
    txid = client.claim_reverse_swap(
        lockup_address,
        receive_address,
        privkey_wif,
        preimage_hex,
        redeem_script_hex,
    )
    click.echo(f"TXID: {txid}")


@click.command()
@click.argument('id', type=str)
def swap_status(id):
    """
    get swap status
    retrieves the status of your boltz swap from the api

    ID is the id of your boltz swap
    """
    client = BoltzClient(config)
    data = client.swap_status(id)
    click.echo(data)


@click.command()
def version():
    """ shows version of boltz api """
    client = BoltzClient(config)
    data = client.check_version()
    click.echo(data["version"])


def main():
    command_group.add_command(swap_status)
    command_group.add_command(create_swap)
    command_group.add_command(refund_swap)
    command_group.add_command(create_reverse_swap)
    command_group.add_command(create_reverse_swap_and_claim)
    command_group.add_command(claim_reverse_swap)
    command_group.add_command(version)
    command_group()


if __name__ == "__main__":
    main()



import os

from typing import Optional
from hashlib import sha256
from binascii import unhexlify, hexlify

from embit import ec, script
from embit.networks import NETWORKS
from embit.transaction import SIGHASH, Transaction, TransactionInput, TransactionOutput

from .mempool import LockupData


def create_preimage() -> tuple[str, str]:
    preimage = os.urandom(32)
    preimage_hash = sha256(preimage).hexdigest()
    return preimage.hex(), preimage_hash


def create_key_pair(network) -> tuple[str, str]:
    net = NETWORKS[network]
    privkey = ec.PrivateKey(os.urandom(32), True, net)
    pubkey_hex = hexlify(privkey.sec()).decode("UTF-8")
    privkey_wif = privkey.wif(net)
    return privkey_wif, pubkey_hex


def create_refund_tx(
    privkey_wif: str,
    receive_address: str,
    redeem_script_hex: str,
    timeout_block_height: int,
    lockup_tx: LockupData,
) -> tuple[str, str]:
    # encrypt redeemscript to script_sig
    rs = bytes([34]) + bytes([0]) + bytes([32])
    rs += sha256(unhexlify(redeem_script_hex)).digest()
    script_sig =  script.Script(data=rs)
    return create_onchain_tx(
        sequence=0xFFFFFFFE,
        redeem_script_hex=redeem_script_hex,
        privkey_wif=privkey_wif,
        lockup_tx=lockup_tx,
        receive_address=receive_address,
        timeout_block_height=timeout_block_height,
        script_sig=script_sig,
    )


def create_claim_tx(
    preimage_hex: str,
    privkey_wif: str,
    receive_address: str,
    redeem_script_hex: str,
    lockup_tx: LockupData
) -> tuple[str, str]:
    return create_onchain_tx(
        preimage_hex=preimage_hex,
        lockup_tx=lockup_tx,
        receive_address=receive_address,
        privkey_wif=privkey_wif,
        redeem_script_hex=redeem_script_hex,
    )


def create_onchain_tx(
    lockup_tx: LockupData,
    receive_address: str,
    privkey_wif: str,
    redeem_script_hex: str,
    fees: int = 1000,
    sequence: int = 0xFFFFFFFF,
    timeout_block_height: int = 0,
    preimage_hex: str = "",
    script_sig: Optional[script.Script]= None,
) -> tuple[str, str]:

    vin = TransactionInput(unhexlify(lockup_tx.txid), lockup_tx.vout_cnt, sequence=sequence)
    vout = TransactionOutput(lockup_tx.vout_amount - fees, script.address_to_scriptpubkey(receive_address))
    tx = Transaction(vin=[vin], vout=[vout])

    if timeout_block_height > 0:
        tx.locktime = timeout_block_height

    if script_sig:
        tx.vin[0].script_sig = script_sig

    # hashing redeemscript
    s = script.Script(data=unhexlify(redeem_script_hex))
    h = tx.sighash_segwit(0, s, lockup_tx.vout_amount)

    # sign the redeemscript hash
    privkey = ec.PrivateKey.from_wif(privkey_wif)
    sig = privkey.sign(h).serialize() + bytes([SIGHASH.ALL])

    # put the witness into the input
    witness_items = [sig, unhexlify(preimage_hex), unhexlify(redeem_script_hex)]
    tx.vin[0].witness = script.Witness(items=witness_items)

    return hexlify(tx.txid()).decode("UTF-8"), hexlify(tx.serialize()).decode("UTF-8")



import json
import httpx
import asyncio
import websockets
import logging
logger = logging.getLogger()

from typing import Optional
from dataclasses import dataclass

from .helpers import req_wrap


@dataclass
class LockupData:
    txid: str
    vout_cnt: int
    vout_amount: int


class MempoolApiException(Exception):
    pass


class MempoolBlockHeightException(Exception):
    pass


class MempoolClient:
    def __init__(self, url, ws_url):
        self._api_url = url
        self._ws_url = ws_url
        # just check if mempool is available
        self.get_blockheight()


    def request(self, funcname, *args, **kwargs) -> dict:
        try:
            return req_wrap(funcname, *args, **kwargs)
        except httpx.RequestError as exc:
            msg = f"unreachable: {exc.request.url!r}."
            raise MempoolApiException(f"mempool api connection error: {msg}")
        except httpx.HTTPStatusError as exc:
            msg = f"{exc.response.status_code} while requesting {exc.request.url!r}. message: {exc.response.text}"
            raise MempoolApiException(f"mempool api status error: {msg}")


    async def wait_for_websocket_message(self, send, message_key):
        async for websocket in websockets.connect(self._ws_url): #type: ignore
            try:
                # await websocket.send(json.dumps({"action": "init"}))
                # await websocket.send(json.dumps({"action": "want", "data": ["blocks", "mempool-blocks"]}))
                await websocket.send(json.dumps({"action": "want", "data": ["blocks"]}))
                await websocket.send(json.dumps(send))
                async for raw in websocket:
                    message = json.loads(raw)
                    if message_key in message:
                        return message.get(message_key)
            except websockets.ConnectionClosed: #type: ignore
                continue


    async def wait_for_one_websocket_message(self, send):
        async with websockets.connect(self._ws_url) as websocket: #type: ignore
            await websocket.send(json.dumps({"action": "want", "data": ["blocks", "mempool-blocks"]}))
            await websocket.send(json.dumps(send))
            raw = await asyncio.wait_for(websocket.recv(), timeout=10)
            return json.loads(raw) if raw else None


    async def wait_for_tx_confirmed(self, txid: str):
        return await self.wait_for_websocket_message({"track-tx": txid }, "txConfirmed")


    async def wait_for_lockup_tx(self, address: str) -> LockupData:
        message = await self.wait_for_websocket_message({"track-address": address }, "address-transactions")
        if not message:
            # restart
            return await self.wait_for_lockup_tx(address)
        lockup_tx = self.find_tx_and_output(message, address)
        if not lockup_tx:
            # restart
            return await self.wait_for_lockup_tx(address)
        return lockup_tx


    def find_tx_and_output(self, txs, address: str) -> Optional[LockupData]:
        if len(txs) == 0:
            return None
        for tx in txs:
            for i, vout in enumerate(tx["vout"]):
                if vout["scriptpubkey_address"] == address:
                    return LockupData(txid=tx["txid"], vout_cnt=i, vout_amount=vout["value"])
        return None


    async def get_tx_from_address(self, address: str) -> LockupData:
        txs = self.request(
            "get",
            f"{self._api_url}/api/address/{address}/txs",
            headers={"Content-Type": "application/json"},
        )
        if len(txs) == 0:
            return await self.wait_for_lockup_tx(address)
        lockup_tx = self.find_tx_and_output(txs, address)
        if not lockup_tx:
            return await self.wait_for_lockup_tx(address)
        return lockup_tx


    def get_fee_estimation(self) -> int:
        # TODO: hardcoded maximum tx size, in the future we try to get the size of the tx via embit
        # we need a function like Transaction.vsize()
        tx_size_vbyte = 200
        mempool_fees = self.get_fees()
        return mempool_fees * tx_size_vbyte


    def get_fees(self) -> int:
        data = self.request(
            "get",
            f"{self._api_url}/api/v1/fees/recommended",
            headers={"Content-Type": "application/json"},
        )
        return int(data["economyFee"])


    def get_blockheight(self) -> int:
        data = self.request(
            "get",
            f"{self._api_url}/api/blocks/tip/height",
            headers={"Content-Type": "text/plain"},
        )
        return int(data["text"])


    def check_block_height(self, timeout_block_height: int) -> None:
        current_block_height = self.get_blockheight()
        if current_block_height < timeout_block_height:
            msg = f"current_block_height ({current_block_height}) has not yet exceeded ({timeout_block_height})"
            raise MempoolBlockHeightException(msg)


    def send_onchain_tx(self, tx_hex: str):
        return self.request(
            "post",
            f"{self._api_url}/api/tx",
            headers={"Content-Type": "text/plain"},
            content=tx_hex,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['boltz_client']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'embit>=0.6.1,<0.7.0',
 'httpx>=0.23.1,<0.24.0',
 'websockets>=10.4,<11.0']

entry_points = \
{'console_scripts': ['boltz = boltz_client.cli:main']}

setup_kwargs = {
    'name': 'boltz-client',
    'version': '0.0.4',
    'description': 'python boltz client',
    'long_description': '# boltz.exchange python client\nA boltz.exchange client for python\n\n## installing\n```console\npoetry install\n```\n\n## running cli\n```console\npoetry run boltz\n```\n\n## starting regtest\n```console\ncd docker\nchmod +x regtest\n./regtest\n```\n\n## running tests\n```console\npoetry run pytest\n```\n',
    'author': 'dni',
    'author_email': 'office@dnilabs.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://boltz.exchange',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
